#!/usr/bin/env python


from datetime import date

__author__ = "Sven Siegmund"
__author_email__ = "sven.siegmund@gmail.com"
__maintainer__ = __author__
__maintainer_email__ = __author_email__
__release_date__ = date(year=2022, month=11, day=13)
__version__ = "0.2.4"
